from travel_backpack.time import time_now_to_string
import functools


def log_info(msg, file, print_to_console=True, print_time=True):
    nmsg = '[' + time_now_to_string() + '] ' + msg
    if print_to_console:
        if print_time:
            print(nmsg)
        else:
            print(msg)
    with open(file, "a+") as f:
        f.write(nmsg + "\n")


def multi_replace(text, dic):
    for i, j in dic.items():
        text = text.replace(i, j)
    return text


class Logger(object):
    def __init__(self,
                 logpath=None,
                 logpath_last=None,
                 timestamp_log=True,
                 timestamp_terminal=True,
                 timestamp_func=None):
        import os
        import sys
        if timestamp_func is None:
            import time
            timestamp_func = lambda: f'[{time.strftime("%d/%m/%Y")} at {time.strftime("%H:%M:%S")}] '
            timestamp_func = lambda: f"[{time_now_to_string(separators=['/','/',' at ',':',':'])}] "

        if logpath is None:
            logpath = os.path.join(os.environ['userprofile'], os.path.basename(__file__) + '.log')

        if logpath_last is None:
            logpath_last = '.last_run'.join(os.path.splitext(logpath))

        self.terminal = sys.stdout
        self.log = open(logpath, "a+", 1)
        self.log_last = open(logpath_last, "w", 1)
        self.timestamp_log = timestamp_log
        self.timestamp_terminal = timestamp_terminal
        self.timestamp_func = timestamp_func

    def write(self, message):
        # Write to terminal
        if self.terminal:
            to_write_terminal = message
            if self.timestamp_terminal:
                to_write_terminal = to_write_terminal.replace('\n', f'\n{self.timestamp_func()}')

            self.terminal.write(to_write_terminal)

        # Write to log
        to_write_log = message
        if self.timestamp_log:
            to_write_log = to_write_log.replace('\n', f'\n{self.timestamp_func()}')

        self.log.write(to_write_log)
        self.log_last.write(to_write_log)

    def flush(self):
        #this flush method is needed for python 3 compatibility.
        #this handles the flush command by doing nothing.
        #you might want to specify some extra behavior here.
        self.log.flush()
        pass


def log_stdout(log_file):
    import sys
    log = Logger(log_file)
    sys.stdout = log
    sys.stderr = log


_log_wrapper_recursion = [None]


def log_wrapper(f):

    f_name = f.__code__.co_name

    @functools.wraps(f)
    def _log_wrapper(*args, **kwargs):

        if not _log_wrapper_recursion[0]:
            # f_name = f.__name__
            _log_wrapper_recursion[0] = f_name
            try:
                f_args = ', '.join(map(str, args))
                f_kwargs = ', '.join([f'{k}={v}' for k, v in kwargs.items()])
                f_cargs = f_args
                if f_kwargs:
                    f_cargs += ', ' + f_kwargs

                f_call = f'{f_name}({f_cargs})'
            except:
                f_call = f'{f_name}(?<exception>?)'
            _log_wrapper_recursion[0] = None
        else:
            f_call = f'{f_name}(?{_log_wrapper_recursion[0]}?)'

        print('=== [log][trace]', '>>>', f_call)

        try:
            result = f(*args, **kwargs)

            if not _log_wrapper_recursion[0]:
                _log_wrapper_recursion[0] = f_name
                try:
                    f_result = f'{result} from {f_call}'
                except:
                    f_result = f'(?<exception>?) from {f_call}'
                _log_wrapper_recursion[0] = None
            else:
                f_result = f'(?{_log_wrapper_recursion[0]}?) from {f_call}'
            print('=== [log][trace]', '<<<', f_result)
        except Exception as ex:
            print('=== [log][trace]', '<<<exception<<<', f'({type(ex)} {ex})', 'from', f_call)
            raise

        return result

    return _log_wrapper


def setup_tracer(name,
                 log_decorator=log_wrapper,
                 dont_log_filter=lambda name: name.startswith('__') and not name.endswith('__')):
    import travel_backpack
    import travel_backpack.decorators
    import sys
    names_to_ignore = [log_decorator.__code__.co_name, log_decorator(lambda x: x).__code__.co_name]

    def log(f):
        if getattr(f, '__code__', None) != None:
            f_name = f.__code__.co_name
            if f_name not in names_to_ignore and not dont_log_filter(f_name):
                # if hasattr(f, '__wrapped__'):
                #     f_child = f.__wrapped__
                #     wrapped_f_child = log(f_child)
                #     f.__wrapped__ = wrapped_f_child

                if travel_backpack.is_local_code(f, name=name):
                    # print('wrapping', f_name, f)
                    return log_decorator(f)

                else:
                    # print('not local', f)
                    return f

            else:
                # print('func already wrapped')
                return f
        else:
            # print('weird function')
            return f

    decorate_methods = travel_backpack.decorators.decorate_all_methods
    decorate_functions = travel_backpack.decorators.decorate_all_functions_in_module
    decorate_classes = travel_backpack.decorators.decorate_all_classes_in_module
    decorate_modules = travel_backpack.decorators.decorate_all_modules_in_module

    def hyper_logger_decorator(function_decorator, class_decorator):
        decorated_modules = set()

        def wrapper(module):
            if module not in decorated_modules:
                decorated_modules.add(module)
                decorate_functions(module, function_decorator)
                decorate_classes(module, class_decorator)
                decorate_modules(module, wrapper, name)
            # else:
            #     print('already decorated:', module)

            return module

        return wrapper

    hyper_logger_decorator(log, decorate_methods(log))(sys.modules[name])


from dataclasses import dataclass
from travel_backpack.exceptions import check_and_raise
from typing import Any, Optional, Sequence, TypeVar, Callable, Type, Union, Tuple, cast, overload, NoReturn
T = TypeVar('T')


@dataclass
class VariableReferenceHolder:
    """Holds a reference to a variable

    Useful for passing as argument just like a
    variable pointer or a variable reference in other languages
    """
    value: Any


class Undefined:
    pass


@overload
def ensure_type(obj: Any,
                _type: Type[T],
                output_type: None = None,
                exception: Union[BaseException, Type[BaseException]] = TypeError) -> T:
    ...

@overload
def ensure_type(obj: Any,
                _type:Union[Type, Tuple[Union[Type, None], ...]],
                output_type: Type[T],
                exception: Union[BaseException, Type[BaseException]] = TypeError) -> T:
    ...


def ensure_type(obj: Any,
                _type: Union[Type[T], Union[Type, Tuple[Union[Type, None], ...]]],
                output_type: Optional[Type[T]] = None,
                exception: Union[BaseException, Type[BaseException]] = TypeError) -> T:
    # if output_type is None:
    #     if isinstance(_type, Tuple):
    #         raise ValueError('output_type must be set when more than one type is given')
    #     else:
    #         output_type = _type

    if isinstance(_type, tuple) and None in _type:
        if obj is None:
            return None
        _type = tuple(t for t in _type if t is not None)

    if isinstance(exception, BaseException):  # is instance of exception
        check_and_raise(isinstance(obj, _type), exception=exception)
    else:  # is class
        exception_message = f"Object of type {type(obj)} does not match type {_type}"
        check_and_raise(isinstance(obj, _type), exception=exception(exception_message))

    return cast(T, obj)


def check_var_input(output_type: Type[T],
                    message: str,
                    type_error_message: str = 'Invalid type',
                    type_constructor: Optional[Callable[[str], T]] = None,
                    end: str = '\n',
                    question_start: str = ' -> ',
                    confirm_answer: bool = False,
                    default: Union[T, Type[Undefined]] = Undefined,
                    show_default: bool = True,
                    **confirmation_binary_user_question_kwargs) -> T:

    input_message = message + end
    if show_default and default != Undefined:
        input_message += '(' + str(default) + ')'

    input_message += question_start

    if type_constructor is None:
        _type_constructor = cast(Callable[[str], T], output_type)
    else:
        _type_constructor = cast(Callable[[str], T], type_constructor)

    while True:
        raw_str_input = input(input_message)

        # if answer is empty and default is set: ans = default
        if raw_str_input == '' and (default is not Undefined):
            result = cast(T, default)

        else:
            try:
                result = _type_constructor(raw_str_input)
                ensure_type(result, output_type)
            except:
                print(type_error_message)
                continue

        if confirm_answer:
            confirm_message_set = 'message' in confirmation_binary_user_question_kwargs
            if not confirm_message_set:
                confirmation_binary_user_question_kwargs['message'] = f'is {result} correct?'

            if binary_user_question(**confirmation_binary_user_question_kwargs):
                break
            else:
                continue
        else:
            break

    return cast(T, result)




from typing import List, Union
time_to_string = lambda x: '%04d/%02d/%02d - %02d:%02d:%02d' % (x.year, x.month, x.day, x.hour, x.minute, x.second)


def time_now_to_string(separators: List[str] = None,
                       order: List[str] = None,
                       lengths: Union[List[int], int] = None) -> str:
    '''
    defaults:

    order = ['y', 'mo', 'd', 'h', 'mi', 's']
    
    separators = ['/', '/', ' - ', ':', ':', '.']
    
    lengths = [4, 2, 2, 2, 2, 2, 4]

    Order can also be changed to show less data. If so, separators must be changed accordingly
    '''
    if order is None:
        order = ['y', 'mo', 'd', 'h', 'mi', 's']
    if separators is None:
        separators = ['/', '/', ' - ', ':', ':', '.']
    if lengths is None:
        lengths = [4, 2, 2, 2, 2, 2, 4]
    elif type(lengths) is int:
        lengths = [lengths] * len(order)

    if len(separators) < len(order) - 1:
        raise Exception('Separator count must be at least order count - 1')
    if len(lengths) < len(order):
        raise Exception('Lengths count must be at least order count or a single int')

    from datetime import datetime as dt
    x = dt.now()
    y = x.year
    m = x.month
    d = x.day
    h = x.hour
    mi = x.minute
    s = x.second
    us = x.microsecond

    var_map = {'y': y, 'mo': m, 'd': d, 'h': h, 'mi': mi, 's': s, 'u': us}
    result = ''
    for i, e in enumerate(order):
        if i > 0:
            result += separators[i - 1]
        result += f'{{0:0{lengths[i]}d}}'.format(var_map[e])
    return result


def format_date_to_path(date):
    return date.replace('/', '-').replace(':', 'h').replace('_', ' ')



import inspect
import functools
import os
from typing import Callable, Any, TypeVar
T = TypeVar('T')
U = TypeVar('U')


def decorate_all_methods(decorator, ignore=['__repr__', '__str__']):
    def decorate(cls):
        is_static = lambda attr: isinstance(inspect.getattr_static(cls, attr), (classmethod, staticmethod))
        is_callable = lambda attr: callable(getattr(cls, attr))
        not_to_ignore = lambda attr: attr not in ignore
        filter_func = lambda attr: not_to_ignore(attr) and is_callable(attr) and not is_static(attr)
        for attr in filter(filter_func, cls.__dict__):  # there's propably a better way to do this
            # print('setting method', attr)
            setattr(cls, attr, decorator(getattr(cls, attr)))

        # print(cls, type(cls), 'callable:', callable(cls))
        return cls

    return decorate


def decorate_all_classes_in_module(module, decorator):
    objects = [getattr(module, name) for name in dir(module)]
    # filter_func = lambda obj: inspect.isclass(obj)
    for obj in filter(inspect.isclass, objects):
        decorator(obj)

    return module


def decorate_all_functions_in_module(module, decorator):
    name_obj_pairs = [(name, getattr(module, name)) for name in dir(module)]
    filter_func = lambda pair: callable(pair[1]) and not inspect.isclass(pair[1])
    for name, obj in filter(filter_func, name_obj_pairs):
        setattr(module, name, decorator(obj))

    return module


def decorate_all_modules_in_module(module, decorator, topmost_path=None, only_local_code=True):
    name_obj_pairs = [(name, getattr(module, name)) for name in dir(module)]
    filter_func = lambda pair: inspect.ismodule(pair[1])

    topmost_dir = ''
    if only_local_code:
        if topmost_path is None:
            topmost_dir: str = os.path.dirname(os.path.realpath(module))
        else:
            topmost_dir: str = os.path.dirname(os.path.realpath(topmost_path))

    for name, obj in filter(filter_func, name_obj_pairs):
        if only_local_code:
            try:
                module_path: str = os.path.realpath(inspect.getabsfile(obj))
                if topmost_dir in module_path:
                    # print('decorating module', name)
                    setattr(module, name, decorator(obj))
                else:
                    # print('ignoring out-of-scope module', name)
                    pass
            except NotImplementedError:
                raise
            except Exception as ex:
                print('module is absolute:', obj, ex)
        else:
            setattr(module, name, decorator(obj))

    return module


def pure_property(func: Callable[[U], T]) -> property:
    """A pure property assumes that the result of
    a property will always be the same given a class
    instance. Doing that, the pure property caches
    the result on the first run and returns it in
    the consecutives 

    Arguments:
        func {function} -- The function to be decorated. must have only self as argument

    Returns:
        function descriptor -- the function wrapped in this decorator and the property decorator
    """

    specific_name = f'__pure_{id(func)}'

    @functools.wraps(func)
    def pure_w(self: U) -> T:
        if not hasattr(self, specific_name):
            # print('returning calculated value')
            setattr(self, specific_name, func(self))
        # else:
        #     print('returning cached value')

        return getattr(self, specific_name)

    # p = pure_w
    # pp = property(p)
    # return pp
    return property(pure_w)


from functools import wraps
from typing import Type, Union, TYPE_CHECKING, cast


def format_exception_string(e):
    import traceback
    s = f'Error: {type(e)} {str(e)}\n'
    s += 'Traceback: ' + '\n'.join(traceback.format_list(traceback.extract_tb(e.__traceback__)))
    return s


def except_and_print(func, message=None, raise_exception=False):
    @wraps(func)
    def decorator(*args, **kwargs):
        try:
            return func(*args, **kwargs)

        except Exception as e:
            if message is not None:
                print(message)
            print(format_exception_string(e))
            if raise_exception:
                raise

    return decorator


def except_and_print_with_message(message, *args, **kwargs):
    def eaprint(func, *args, **kwargs):
        return except_and_print(func, *args, message=message, **kwargs)

    return eaprint


def check_and_raise(condition: bool, exception: Union[Type[BaseException], BaseException] = Exception):
    """Raises an exception when condition is not met
    
    Arguments:
        condition {bool} -- condition to check against
    
    Keyword Arguments:
        exception {Union[Type[Exception],Exception]} -- the exception type to be raised (default: {Exception})
    
    Raises:
        exception_type: the exception of type exception_type, given in the parameter
    """
    if not condition:
        if TYPE_CHECKING:
            if isinstance(exception, BaseException):
                exception = cast(BaseException, exception)
                raise exception
            else:
                exception = cast(Type[BaseException], exception)
                raise exception
        else:
            raise exception
